# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the RT-DETR implementation of lyuwenyu.
https://github.com/lyuwenyu/RT-DETR/blob/main/rtdetr_pytorch/src/zoo/rtdetr/rtdetr_decoder.py
"""
from collections import OrderedDict
import copy
import math
from typing import List

from omegaconf import DictConfig
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init

from ....utils import ModelOutput
from ....op.registry import ACTIVATION_REGISTRY
from ....op.denoising import get_contrastive_denoising_training_group


def deformable_attention_core_func(value, value_spatial_shapes, sampling_locations, attention_weights):
    """
    Args:
        value (Tensor): [bs, value_length, n_head, c]
        value_spatial_shapes (Tensor|List): [n_levels, 2]
        value_level_start_index (Tensor|List): [n_levels]
        sampling_locations (Tensor): [bs, query_length, n_head, n_levels, n_points, 2]
        attention_weights (Tensor): [bs, query_length, n_head, n_levels, n_points]

    Returns:
        output (Tensor): [bs, Length_{query}, C]
    """
    bs, _, n_head, c = value.shape
    _, Len_q, _, n_levels, n_points, _ = sampling_locations.shape

    split_shape = [h * w for h, w in value_spatial_shapes]
    value_list = value.split(split_shape, dim=1)
    sampling_grids = 2 * sampling_locations - 1
    sampling_value_list = []
    for level, (h, w) in enumerate(value_spatial_shapes):
        # N_, H_*W_, M_, D_ -> N_, H_*W_, M_*D_ -> N_, M_*D_, H_*W_ -> N_*M_, D_, H_, W_
        value_l_ = value_list[level].flatten(2).permute(
            0, 2, 1).reshape(bs * n_head, c, h, w)
        # N_, Lq_, M_, P_, 2 -> N_, M_, Lq_, P_, 2 -> N_*M_, Lq_, P_, 2
        sampling_grid_l_ = sampling_grids[:, :, :, level].permute(
            0, 2, 1, 3, 4).flatten(0, 1)
        # N_*M_, D_, Lq_, P_
        sampling_value_l_ = F.grid_sample(
            value_l_,
            sampling_grid_l_,
            mode='bilinear',
            padding_mode='zeros',
            align_corners=False)
        sampling_value_list.append(sampling_value_l_)
    # (N_, Lq_, M_, L_, P_) -> (N_, M_, Lq_, L_, P_) -> (N_*M_, 1, Lq_, L_*P_)
    attention_weights = attention_weights.permute(0, 2, 1, 3, 4).reshape(
        bs * n_head, 1, Len_q, n_levels * n_points)
    output = (torch.stack(
        sampling_value_list, dim=-2).flatten(-2) *
              attention_weights).sum(-1).reshape(bs, n_head * c, Len_q)

    return output.permute(0, 2, 1)


def inverse_sigmoid(x: torch.Tensor, eps: float=1e-5) -> torch.Tensor:
    x = x.clip(min=0., max=1.)
    return torch.log(x.clip(min=eps) / (1 - x).clip(min=eps))


def bias_init_with_prob(prior_prob=0.01):
    """initialize conv/fc bias value according to a given probability value."""
    bias_init = float(-math.log((1 - prior_prob) / prior_prob))
    return bias_init


# TODO: Replace with custom implementation
class MLP(nn.Module):
    def __init__(self, input_dim, hidden_dim, output_dim, num_layers, act='relu'):
        super().__init__()
        self.num_layers = num_layers
        h = [hidden_dim] * (num_layers - 1)
        self.layers = nn.ModuleList(nn.Linear(n, k) for n, k in zip([input_dim] + h, h + [output_dim]))
        self.act = nn.Identity() if act is None else ACTIVATION_REGISTRY[act]()

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            x = self.act(layer(x)) if i < self.num_layers - 1 else layer(x)
        return x


# TODO: Replace with custom implementation
class MSDeformableAttention(nn.Module):
    def __init__(self, embed_dim=256, num_heads=8, num_levels=4, num_points=4,):
        """
        Multi-Scale Deformable Attention Module
        """
        super(MSDeformableAttention, self).__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.num_levels = num_levels
        self.num_points = num_points
        self.total_points = num_heads * num_levels * num_points

        self.head_dim = embed_dim // num_heads
        assert self.head_dim * num_heads == self.embed_dim, "embed_dim must be divisible by num_heads"

        self.sampling_offsets = nn.Linear(embed_dim, self.total_points * 2,)
        self.attention_weights = nn.Linear(embed_dim, self.total_points)
        self.value_proj = nn.Linear(embed_dim, embed_dim)
        self.output_proj = nn.Linear(embed_dim, embed_dim)

        self.ms_deformable_attn_core = deformable_attention_core_func

        self._reset_parameters()


    def _reset_parameters(self):
        # sampling_offsets
        init.constant_(self.sampling_offsets.weight, 0)
        thetas = torch.arange(self.num_heads, dtype=torch.float32) * (2.0 * math.pi / self.num_heads)
        grid_init = torch.stack([thetas.cos(), thetas.sin()], -1)
        grid_init = grid_init / grid_init.abs().max(-1, keepdim=True).values
        grid_init = grid_init.reshape(self.num_heads, 1, 1, 2).tile([1, self.num_levels, self.num_points, 1])
        scaling = torch.arange(1, self.num_points + 1, dtype=torch.float32).reshape(1, 1, -1, 1)
        grid_init *= scaling
        self.sampling_offsets.bias.data[...] = grid_init.flatten()

        # attention_weights
        init.constant_(self.attention_weights.weight, 0)
        init.constant_(self.attention_weights.bias, 0)

        # proj
        init.xavier_uniform_(self.value_proj.weight)
        init.constant_(self.value_proj.bias, 0)
        init.xavier_uniform_(self.output_proj.weight)
        init.constant_(self.output_proj.bias, 0)


    def forward(self,
                query,
                reference_points,
                value,
                value_spatial_shapes,
                value_mask=None):
        """
        Args:
            query (Tensor): [bs, query_length, C]
            reference_points (Tensor): [bs, query_length, n_levels, 2], range in [0, 1], top-left (0,0),
                bottom-right (1, 1), including padding area
            value (Tensor): [bs, value_length, C]
            value_spatial_shapes (List): [n_levels, 2], [(H_0, W_0), (H_1, W_1), ..., (H_{L-1}, W_{L-1})]
            value_level_start_index (List): [n_levels], [0, H_0*W_0, H_0*W_0+H_1*W_1, ...]
            value_mask (Tensor): [bs, value_length], True for non-padding elements, False for padding elements

        Returns:
            output (Tensor): [bs, Length_{query}, C]
        """
        bs, Len_q = query.shape[:2]
        Len_v = value.shape[1]

        value = self.value_proj(value)
        if value_mask is not None:
            value_mask = value_mask.astype(value.dtype).unsqueeze(-1)
            value *= value_mask
        value = value.reshape(bs, Len_v, self.num_heads, self.head_dim)

        sampling_offsets = self.sampling_offsets(query).reshape(
            bs, Len_q, self.num_heads, self.num_levels, self.num_points, 2)
        attention_weights = self.attention_weights(query).reshape(
            bs, Len_q, self.num_heads, self.num_levels * self.num_points)
        attention_weights = F.softmax(attention_weights, dim=-1).reshape(
            bs, Len_q, self.num_heads, self.num_levels, self.num_points)

        if reference_points.shape[-1] == 2:
            offset_normalizer = torch.tensor(value_spatial_shapes)
            offset_normalizer = offset_normalizer.flip([1]).reshape(
                1, 1, 1, self.num_levels, 1, 2)
            sampling_locations = reference_points.reshape(
                bs, Len_q, 1, self.num_levels, 1, 2
            ) + sampling_offsets / offset_normalizer
        elif reference_points.shape[-1] == 4:
            sampling_locations = (
                reference_points[:, :, None, :, None, :2] + sampling_offsets /
                self.num_points * reference_points[:, :, None, :, None, 2:] * 0.5)
        else:
            raise ValueError(
                "Last dim of reference_points must be 2 or 4, but get {} instead.".
                format(reference_points.shape[-1]))

        output = self.ms_deformable_attn_core(value, value_spatial_shapes, sampling_locations, attention_weights)

        output = self.output_proj(output)

        return output


class TransformerDecoderLayer(nn.Module):
    def __init__(self,
                 d_model=256,
                 n_head=8,
                 dim_feedforward=1024,
                 dropout=0.,
                 activation="relu",
                 n_levels=4,
                 n_points=4,):
        super(TransformerDecoderLayer, self).__init__()

        # self attention
        self.self_attn = nn.MultiheadAttention(d_model, n_head, dropout=dropout, batch_first=True)
        self.dropout1 = nn.Dropout(dropout)
        self.norm1 = nn.LayerNorm(d_model)

        # cross attention
        self.cross_attn = MSDeformableAttention(d_model, n_head, n_levels, n_points)
        self.dropout2 = nn.Dropout(dropout)
        self.norm2 = nn.LayerNorm(d_model)

        # ffn
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.activation = getattr(F, activation)
        self.dropout3 = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)
        self.dropout4 = nn.Dropout(dropout)
        self.norm3 = nn.LayerNorm(d_model)

        # self._reset_parameters()

    # def _reset_parameters(self):
    #     linear_init_(self.linear1)
    #     linear_init_(self.linear2)
    #     xavier_uniform_(self.linear1.weight)
    #     xavier_uniform_(self.linear2.weight)

    def with_pos_embed(self, tensor, pos):
        return tensor if pos is None else tensor + pos

    def forward_ffn(self, tgt):
        return self.linear2(self.dropout3(self.activation(self.linear1(tgt))))

    def forward(self,
                tgt,
                reference_points,
                memory,
                memory_spatial_shapes,
                memory_level_start_index,
                attn_mask=None,
                memory_mask=None,
                query_pos_embed=None):
        # self attention
        q = k = self.with_pos_embed(tgt, query_pos_embed)

        # if attn_mask is not None:
        #     attn_mask = torch.where(
        #         attn_mask.to(torch.bool),
        #         torch.zeros_like(attn_mask),
        #         torch.full_like(attn_mask, float('-inf'), dtype=tgt.dtype))

        tgt2, _ = self.self_attn(q, k, value=tgt, attn_mask=attn_mask)
        tgt = tgt + self.dropout1(tgt2)
        tgt = self.norm1(tgt)

        # cross attention
        tgt2 = self.cross_attn(\
            self.with_pos_embed(tgt, query_pos_embed), 
            reference_points, 
            memory, 
            memory_spatial_shapes, 
            memory_mask)
        tgt = tgt + self.dropout2(tgt2)
        tgt = self.norm2(tgt)

        # ffn
        tgt2 = self.forward_ffn(tgt)
        tgt = tgt + self.dropout4(tgt2)
        tgt = self.norm3(tgt)

        return tgt


class TransformerDecoder(nn.Module):
    def __init__(self, hidden_dim, decoder_layer, num_layers, eval_idx=-1):
        super(TransformerDecoder, self).__init__()
        self.layers = nn.ModuleList([copy.deepcopy(decoder_layer) for _ in range(num_layers)])
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        self.eval_idx = eval_idx if eval_idx >= 0 else num_layers + eval_idx

    def forward(self,
                tgt,
                ref_points_unact,
                memory,
                memory_spatial_shapes,
                memory_level_start_index,
                bbox_head,
                score_head,
                query_pos_head,
                attn_mask=None,
                memory_mask=None):
        output = tgt
        dec_out_bboxes = []
        dec_out_logits = []
        ref_points_detach = torch.sigmoid(ref_points_unact)

        for i, layer in enumerate(self.layers):
            ref_points_input = ref_points_detach.unsqueeze(2)
            query_pos_embed = query_pos_head(ref_points_detach)

            output = layer(output, ref_points_input, memory,
                           memory_spatial_shapes, memory_level_start_index,
                           attn_mask, memory_mask, query_pos_embed)

            inter_ref_bbox = torch.sigmoid(bbox_head[i](output) + inverse_sigmoid(ref_points_detach))

            if self.training:
                dec_out_logits.append(score_head[i](output))
                if i == 0:
                    dec_out_bboxes.append(inter_ref_bbox)
                else:
                    dec_out_bboxes.append(torch.sigmoid(bbox_head[i](output) + inverse_sigmoid(ref_points)))

            elif i == self.eval_idx:
                dec_out_logits.append(score_head[i](output))
                dec_out_bboxes.append(inter_ref_bbox)
                break

            ref_points = inter_ref_bbox
            ref_points_detach = inter_ref_bbox.detach(
            ) if self.training else inter_ref_bbox

        return torch.stack(dec_out_bboxes), torch.stack(dec_out_logits)


class RTDETRTransformer(nn.Module):
    def __init__(
        self,
        num_classes: int,
        intermediate_features_dim: List[int],
        params: DictConfig,
    ):

        super(RTDETRTransformer, self).__init__()

        self.num_classes = num_classes
        feat_channels = intermediate_features_dim
        self.feat_strides = [8, 16, 32] # TODO: Get from other source

        self.hidden_dim = params.hidden_dim
        self.nhead = params.num_attention_heads
        self.num_levels = params.num_levels
        self.num_queries = params.num_queries
        self.eps = params.eps
        self.num_decoder_layers = params.num_decoder_layers # 6
        self.eval_spatial_size = params.eval_spatial_size
        
        position_embed_type = params.position_embed_type
        num_decoder_points = params.num_decoder_points
        dim_feedforward = params.dim_feedforward
        dropout = params.dropout
        activation = params.act_type

        # TODO: Move to loss module
        self.use_aux_loss = params.use_aux_loss
        num_denoising = params.num_denoising
        label_noise_ratio = params.label_noise_ratio
        box_noise_scale=1.0
        learnt_init_query=False
        eval_idx=-1

        assert position_embed_type in ['sine', 'learned'], \
            f'ValueError: position_embed_type not supported {position_embed_type}!'
        assert len(feat_channels) <= self.num_levels
        assert len(self.feat_strides) == len(feat_channels)
        for _ in range(self.num_levels - len(self.feat_strides)):
            self.feat_strides.append(self.feat_strides[-1] * 2)

        # backbone feature projection
        self._build_input_proj_layer(feat_channels)

        # Transformer module
        decoder_layer = TransformerDecoderLayer(self.hidden_dim, self.nhead, dim_feedforward, dropout, activation, self.num_levels, num_decoder_points)
        self.decoder = TransformerDecoder(self.hidden_dim, decoder_layer, self.num_decoder_layers, eval_idx)

        self.num_denoising = num_denoising
        self.label_noise_ratio = label_noise_ratio
        self.box_noise_scale = box_noise_scale
        # denoising part
        if num_denoising > 0: 
            self._training_targets = None
            # self.denoising_class_embed = nn.Embedding(num_classes, hidden_dim, padding_idx=num_classes-1) # TODO for load paddle weights
            self.denoising_class_embed = nn.Embedding(self.num_classes+1, self.hidden_dim, padding_idx=self.num_classes)
            init.normal_(self.denoising_class_embed.weight[:-1])

        # decoder embedding
        self.learnt_init_query = learnt_init_query
        if learnt_init_query:
            self.tgt_embed = nn.Embedding(self.num_queries, self.hidden_dim)
        self.query_pos_head = MLP(4, 2 * self.hidden_dim, self.hidden_dim, num_layers=2)
        # encoder head

        self.enc_output = nn.Sequential(
            nn.Linear(self.hidden_dim, self.hidden_dim),
            nn.LayerNorm(self.hidden_dim,)
        )
        self.enc_score_head = nn.Linear(self.hidden_dim, self.num_classes)
        self.enc_bbox_head = MLP(self.hidden_dim, self.hidden_dim, 4, num_layers=3)

        # decoder head
        self.dec_score_head = nn.ModuleList([
            nn.Linear(self.hidden_dim, self.num_classes)
            for _ in range(self.num_decoder_layers)
        ])
        self.dec_bbox_head = nn.ModuleList([
            MLP(self.hidden_dim, self.hidden_dim, 4, num_layers=3)
            for _ in range(self.num_decoder_layers)
        ])

        # init encoder output anchors and valid_mask
        if self.eval_spatial_size:
            self.anchors, self.valid_mask = self._generate_anchors()

        self._reset_parameters()

    def _reset_parameters(self):
        bias = bias_init_with_prob(0.01)

        init.constant_(self.enc_score_head.bias, bias)
        init.constant_(self.enc_bbox_head.layers[-1].weight, 0)
        init.constant_(self.enc_bbox_head.layers[-1].bias, 0)

        for cls_, reg_ in zip(self.dec_score_head, self.dec_bbox_head):
            init.constant_(cls_.bias, bias)
            init.constant_(reg_.layers[-1].weight, 0)
            init.constant_(reg_.layers[-1].bias, 0)
        
        # linear_init_(self.enc_output[0])
        init.xavier_uniform_(self.enc_output[0].weight)
        if self.learnt_init_query:
            init.xavier_uniform_(self.tgt_embed.weight)
        init.xavier_uniform_(self.query_pos_head.layers[0].weight)
        init.xavier_uniform_(self.query_pos_head.layers[1].weight)


    def _build_input_proj_layer(self, feat_channels):
        self.input_proj = nn.ModuleList()
        for in_channels in feat_channels:
            self.input_proj.append(
                nn.Sequential(OrderedDict([
                    ('conv', nn.Conv2d(in_channels, self.hidden_dim, 1, bias=False)), 
                    ('norm', nn.BatchNorm2d(self.hidden_dim,))])
                )
            )

        in_channels = feat_channels[-1]

        for _ in range(self.num_levels - len(feat_channels)):
            self.input_proj.append(
                nn.Sequential(OrderedDict([
                    ('conv', nn.Conv2d(in_channels, self.hidden_dim, 3, 2, padding=1, bias=False)),
                    ('norm', nn.BatchNorm2d(self.hidden_dim))])
                )
            )
            in_channels = self.hidden_dim

    def _get_encoder_input(self, feats):
        # get projection features
        proj_feats = [self.input_proj[i](feat) for i, feat in enumerate(feats)]
        if self.num_levels > len(proj_feats):
            len_srcs = len(proj_feats)
            for i in range(len_srcs, self.num_levels):
                if i == len_srcs:
                    proj_feats.append(self.input_proj[i](feats[-1]))
                else:
                    proj_feats.append(self.input_proj[i](proj_feats[-1]))

        # get encoder inputs
        feat_flatten = []
        spatial_shapes = []
        level_start_index = [0, ]
        for i, feat in enumerate(proj_feats):
            _, _, h, w = feat.shape
            # [b, c, h, w] -> [b, h*w, c]
            feat_flatten.append(feat.flatten(2).permute(0, 2, 1))
            # [num_levels, 2]
            spatial_shapes.append([h, w])
            # [l], start index of each level
            level_start_index.append(h * w + level_start_index[-1])

        # [b, l, c]
        feat_flatten = torch.concat(feat_flatten, 1)
        level_start_index.pop()
        return (feat_flatten, spatial_shapes, level_start_index)

    def _generate_anchors(self,
                          spatial_shapes=None,
                          grid_size=0.05,
                          dtype=torch.float32,
                          device='cpu'):
        if spatial_shapes is None:
            spatial_shapes = [[int(self.eval_spatial_size[0] / s), int(self.eval_spatial_size[1] / s)]
                for s in self.feat_strides
            ]
        anchors = []
        for lvl, (h, w) in enumerate(spatial_shapes):
            grid_y, grid_x = torch.meshgrid(\
                torch.arange(end=h, dtype=dtype), \
                torch.arange(end=w, dtype=dtype), indexing='ij')
            grid_xy = torch.stack([grid_x, grid_y], -1)
            valid_WH = torch.tensor([w, h]).to(dtype)
            grid_xy = (grid_xy.unsqueeze(0) + 0.5) / valid_WH
            wh = torch.ones_like(grid_xy) * grid_size * (2.0 ** lvl)
            anchors.append(torch.concat([grid_xy, wh], -1).reshape(-1, h * w, 4))

        anchors = torch.concat(anchors, 1).to(device)
        valid_mask = ((anchors > self.eps) * (anchors < 1 - self.eps)).all(-1, keepdim=True)
        anchors = torch.log(anchors / (1 - anchors))
        # anchors = torch.where(valid_mask, anchors, float('inf'))
        # anchors[valid_mask] = torch.inf # valid_mask [1, 8400, 1]
        anchors = torch.where(valid_mask, anchors, torch.inf)

        return anchors, valid_mask


    def _get_decoder_input(self,
                           memory,
                           spatial_shapes,
                           denoising_class=None,
                           denoising_bbox_unact=None):
        bs, _, _ = memory.shape
        # prepare input for decoder
        if self.training or self.eval_spatial_size is None:
            anchors, valid_mask = self._generate_anchors(spatial_shapes, device=memory.device)
        else:
            anchors, valid_mask = self.anchors.to(memory.device), self.valid_mask.to(memory.device)

        # memory = torch.where(valid_mask, memory, 0)
        memory = valid_mask.to(memory.dtype) * memory  # TODO fix type error for onnx export 

        output_memory = self.enc_output(memory)

        enc_outputs_class = self.enc_score_head(output_memory)
        enc_outputs_coord_unact = self.enc_bbox_head(output_memory) + anchors

        _, topk_ind = torch.topk(enc_outputs_class.max(-1).values, self.num_queries, dim=1)
        
        reference_points_unact = enc_outputs_coord_unact.gather(dim=1, \
            index=topk_ind.unsqueeze(-1).repeat(1, 1, enc_outputs_coord_unact.shape[-1]))

        enc_topk_bboxes = torch.sigmoid(reference_points_unact)
        if denoising_bbox_unact is not None:
            reference_points_unact = torch.concat(
                [denoising_bbox_unact, reference_points_unact], 1)
        
        enc_topk_logits = enc_outputs_class.gather(dim=1, \
            index=topk_ind.unsqueeze(-1).repeat(1, 1, enc_outputs_class.shape[-1]))

        # extract region features
        if self.learnt_init_query:
            target = self.tgt_embed.weight.unsqueeze(0).tile([bs, 1, 1])
        else:
            target = output_memory.gather(dim=1, \
                index=topk_ind.unsqueeze(-1).repeat(1, 1, output_memory.shape[-1]))
            target = target.detach()

        if denoising_class is not None:
            target = torch.concat([denoising_class, target], 1)

        return target, reference_points_unact.detach(), enc_topk_bboxes, enc_topk_logits


    def forward(self, feats):

        # input projection and embedding
        (memory, spatial_shapes, level_start_index) = self._get_encoder_input(feats)

        # prepare denoising training
        if self.training and self.num_denoising > 0:
            targets = self._training_targets
            assert targets is not None
            img_size = targets['img_size']
            targets = targets['gt']
            # Normalize bounding boxes
            gt_for_denoising = list()
            for idx, t in enumerate(targets):
                temp_gt = dict()
                normalized_bboxes = self.normalize_bboxes(t['boxes'], img_size)
                temp_gt['boxes'] = normalized_bboxes
                temp_gt['labels'] = t['labels']
                gt_for_denoising.append(temp_gt)
            denoising_class, denoising_bbox_unact, attn_mask, dn_meta = \
                get_contrastive_denoising_training_group(gt_for_denoising, \
                    self.num_classes, 
                    self.num_queries, 
                    self.denoising_class_embed, 
                    num_denoising=self.num_denoising, 
                    label_noise_ratio=self.label_noise_ratio, 
                    box_noise_scale=self.box_noise_scale, )
        else:
            denoising_class, denoising_bbox_unact, attn_mask, dn_meta = None, None, None, None
        target, init_ref_points_unact, enc_topk_bboxes, enc_topk_logits = \
            self._get_decoder_input(memory, spatial_shapes, denoising_class, denoising_bbox_unact)

        # decoder
        out_bboxes, out_logits = self.decoder(
            target,
            init_ref_points_unact,
            memory,
            spatial_shapes,
            level_start_index,
            self.dec_bbox_head,
            self.dec_score_head,
            self.query_pos_head,
            attn_mask=attn_mask)

        if self.training and dn_meta is not None:
            dn_out_bboxes, out_bboxes = torch.split(out_bboxes, dn_meta['dn_num_split'], dim=2)
            dn_out_logits, out_logits = torch.split(out_logits, dn_meta['dn_num_split'], dim=2)
            self.clear_training_targets()

        out = torch.cat([out_bboxes[-1], out_logits[-1]], dim=-1)

        if self.training and self.use_aux_loss:
            aux_outputs = self._set_aux_loss(out_logits[:-1], out_bboxes[:-1])
            aux_outputs.extend(self._set_aux_loss([enc_topk_logits], [enc_topk_bboxes]))
            if self.training and dn_meta is not None:
                dn_aux_outputs = self._set_aux_loss(dn_out_logits, dn_out_bboxes)
            else:
                dn_aux_outputs = None
        else:
            aux_outputs = None
            dn_aux_outputs = None
        
        return ModelOutput(pred=out, aux_pred=aux_outputs, dn_aux_pred=dn_aux_outputs, dn_meta=dn_meta)

    def _set_aux_loss(self, outputs_class, outputs_coord):
        return [{'pred_logits': a, 'pred_boxes': b}
                for a, b in zip(outputs_class, outputs_coord)]
    
    def normalize_bboxes(self, bboxes, img_size):
        """ Normalize bounding boxes to [0, 1] range """
        if isinstance(img_size, int):
            h = w = img_size
        else:
            h, w = img_size
        scale = torch.tensor([w, h, w, h], dtype=bboxes.dtype, device=bboxes.device)
        return bboxes / scale.unsqueeze(0)

    def set_training_targets(self, targets):
        """Set internal training targets before training step"""
        self._training_targets = targets

    def clear_training_targets(self):
        """Clear internal training targets after training step"""
        self._training_targets = None

def rtdetr_head(num_classes, intermediate_features_dim, conf_model_head, **kwargs) -> RTDETRTransformer:
    return RTDETRTransformer(num_classes=num_classes,
                             intermediate_features_dim=intermediate_features_dim,
                             params=conf_model_head.params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------
from typing import Optional

import torch

from ..models.utils import ModelOutput

TOPK_MAX = 20


class ClassificationPostprocessor():
    def __init__(self, conf_model):
        pass

    def __call__(self, outputs: ModelOutput, k: Optional[int]=None):
        pred = outputs['pred']
        maxk = min(TOPK_MAX, pred.size()[1])
        if k:
            maxk = min(k, maxk)
        logits, pred = pred.topk(maxk, 1, True, True)
        conf_score = torch.softmax(logits, dim=-1)
        return pred.detach().cpu().numpy(), conf_score.detach().cpu().numpy()



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from ctypes import c_int
from multiprocessing import Value
from pathlib import Path
from typing import Dict, Union

import torch
import torch.distributed as dist
import torch.nn as nn
from loguru import logger
from omegaconf import DictConfig
from torch.utils.data import DataLoader

from ..loggers import build_logger
from ..losses import build_losses
from ..metrics import build_metrics
from ..optimizers import build_optimizer
from ..postprocessors import build_postprocessor
from ..schedulers import build_scheduler
from ..utils.model_ema import build_ema
from ..utils.record import Timer
from .registry import PIPELINES, SUPPORTING_TASK_LIST, TASK_PROCESSOR
from .train import NUM_SAMPLES


def load_optimizer_checkpoint(conf, optimizer, scheduler):
    start_epoch = 1
    resume_optimizer_checkpoint = conf.model.checkpoint.optimizer_path
    if resume_optimizer_checkpoint is not None:
        resume_optimizer_checkpoint = Path(resume_optimizer_checkpoint)
        if not resume_optimizer_checkpoint.exists():
            logger.warning(f"Traning summary checkpoint path {str(resume_optimizer_checkpoint)} is not found!"
                            f"Skip loading the previous history and trainer will be started from the beginning")
            return

        optimizer_dict = torch.load(resume_optimizer_checkpoint, map_location='cpu')
        optimizer_state_dict = optimizer_dict['optimizer']
        start_epoch = optimizer_dict['last_epoch'] + 1  # Start from the next to the end of last training

        optimizer.load_state_dict(optimizer_state_dict)
        scheduler.step(epoch=start_epoch)

        start_epoch = start_epoch
        logger.info(f"Resume training from {str(resume_optimizer_checkpoint)}. Start training at epoch: {start_epoch}")

    return optimizer, scheduler, start_epoch


def build_pipeline(
    pipeline_type: str,
    conf: DictConfig,
    task: str,
    model_name: str,
    model: nn.Module,
    devices: torch.device,
    class_map: Dict,
    logging_dir: Union[str, Path],
    is_graphmodule_training: bool,
    dataloaders: Dict[str, DataLoader],
    data_stats: Dict[str, Dict],
    profile: bool = False
):
    assert task in SUPPORTING_TASK_LIST, f"No such task! (task: {task})"

    # Build task processor
    postprocessor = build_postprocessor(task, conf.model)
    task_processor = TASK_PROCESSOR[task](conf, postprocessor, devices, num_classes=len(class_map))

    # Build timer
    timer = Timer()

    if pipeline_type == 'train':
        train_dataloader: DataLoader = dataloaders['train']
        eval_dataloader: DataLoader = dataloaders['valid']

        train_data_stats: Dict = data_stats['train']
        eval_data_stats: Dict = data_stats['valid']
        # Build optimizer and scheduler
        optimizer = build_optimizer(model, single_task_model=conf.model.single_task_model, optimizer_conf=conf.training.optimizer)
        scheduler, _ = build_scheduler(optimizer, conf.training)
        optimizer, scheduler, start_epoch = load_optimizer_checkpoint(conf, optimizer, scheduler)

        # Set current epoch counter and end epoch in dataloader.dataset to use in dataset.transforms
        cur_epoch = Value(c_int, start_epoch)
        train_dataloader.dataset.cur_epoch = cur_epoch
        train_dataloader.dataset.end_epoch = conf.training.epochs

        # Build loss and metric modules
        loss_factory = build_losses(conf.model, cur_epoch=cur_epoch)
        metric_factory = build_metrics(task, conf.model, conf.logging.metrics, train_dataloader.dataset.num_classes)

        # Set model EMA
        model_ema = None
        if conf.training.ema:
            model_ema = build_ema(model=model.module if hasattr(model, 'module') else model, conf=conf)

        # Build logger
        single_gpu_or_rank_zero = (not conf.distributed) or (conf.distributed and dist.get_rank() == 0)
        train_step_per_epoch = len(train_dataloader)
        train_logger = None
        if single_gpu_or_rank_zero:
            train_logger = build_logger(conf, task, model_name,
                                        step_per_epoch=train_step_per_epoch,
                                        class_map=class_map,
                                        num_sample_images=NUM_SAMPLES,
                                        result_dir=logging_dir,)

        # Build pipeline
        pipeline = PIPELINES[pipeline_type](conf=conf,
                                            task=task,
                                            task_processor=task_processor,
                                            model_name=model_name,
                                            model=model,
                                            logger=train_logger,
                                            timer=timer,
                                            optimizer=optimizer,
                                            scheduler=scheduler,
                                            loss_factory=loss_factory,
                                            metric_factory=metric_factory,
                                            train_dataloader=train_dataloader,
                                            eval_dataloader=eval_dataloader,
                                            train_data_stats=train_data_stats,
                                            eval_data_stats=eval_data_stats,
                                            single_gpu_or_rank_zero=single_gpu_or_rank_zero,
                                            is_graphmodule_training=is_graphmodule_training,
                                            model_ema=model_ema,
                                            start_epoch=start_epoch,
                                            cur_epoch=cur_epoch,
                                            profile=profile)

    elif pipeline_type == 'evaluation':
        eval_dataloader: DataLoader = dataloaders['test']
        eval_data_stats: Dict = data_stats['test']

        # Build modules for evaluation
        loss_factory = build_losses(conf.model)
        metric_factory = build_metrics(task, conf.model, conf.logging.metrics, eval_dataloader.dataset.num_classes)

        # Build logger
        single_gpu_or_rank_zero = (not conf.distributed) or (conf.distributed and dist.get_rank() == 0)
        eval_logger = None
        if single_gpu_or_rank_zero:
            eval_logger = build_logger(conf, task, model_name,
                                       step_per_epoch=0,
                                       class_map=class_map,
                                       num_sample_images=NUM_SAMPLES,
                                       result_dir=logging_dir,)
        # Build pipeline
        pipeline = PIPELINES[pipeline_type](conf=conf,
                                            task=task,
                                            task_processor=task_processor,
                                            model_name=model_name,
                                            model=model,
                                            logger=eval_logger,
                                            timer=timer,
                                            loss_factory=loss_factory,
                                            metric_factory=metric_factory,
                                            eval_dataloader=eval_dataloader,
                                            eval_data_stats=eval_data_stats,
                                            single_gpu_or_rank_zero=single_gpu_or_rank_zero,)

    elif pipeline_type == 'inference':
        test_dataloader: DataLoader = dataloaders['test']

        # Build logger
        single_gpu_or_rank_zero = (not conf.distributed) or (conf.distributed and dist.get_rank() == 0)
        eval_logger = None
        if single_gpu_or_rank_zero:
            eval_logger = build_logger(conf, task, model_name,
                                       step_per_epoch=0,
                                       class_map=class_map,
                                       num_sample_images=NUM_SAMPLES,
                                       result_dir=logging_dir,)
        # Build pipeline
        pipeline = PIPELINES[pipeline_type](conf=conf,
                                            task=task,
                                            task_processor=task_processor,
                                            model_name=model_name,
                                            model=model,
                                            logger=eval_logger,
                                            timer=timer,
                                            test_dataloader=test_dataloader,
                                            single_gpu_or_rank_zero=single_gpu_or_rank_zero,)

    else:
        raise ValueError("``pipeline_type`` must be one of ['train', 'evaluation', 'inference'].")

    return pipeline



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import json
import os
from functools import partial
from itertools import chain
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import PIL.Image as Image
import torch
import torch.distributed as dist
from loguru import logger
from omegaconf import DictConfig, ListConfig

from .augmentation.transforms import generate_edge
from .base import BaseCustomDataset, BaseHFDataset, BaseSampleLoader
from .utils.constants import IMG_EXTENSIONS
from .utils.misc import as_tuple, natural_key


class SegmentationSampleLoader(BaseSampleLoader):
    def __init__(self, conf_data, train_valid_split_ratio):
        super(SegmentationSampleLoader, self).__init__(conf_data, train_valid_split_ratio)

    def load_data(self, split='train'):
        assert split in ['train', 'valid', 'test'], f"split should be either {['train', 'valid', 'test']}."
        data_root = Path(self.conf_data.path.root)
        split_dir = self.conf_data.path[split]
        image_dir: Path = data_root / split_dir.image
        annotation_dir: Optional[Path] = data_root / split_dir.label if split_dir.label is not None else None
        images: List[str] = []
        labels: List[str] = []
        images_and_targets: List[Dict[str, str]] = []
        if annotation_dir is not None:
            for ext in IMG_EXTENSIONS:
                images.extend([str(file) for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}'))])
                # TODO: get paired data from regex pattern matching (conf_data.path.pattern)
                labels.extend([str(file) for file in chain(annotation_dir.glob(f'*{ext}'), annotation_dir.glob(f'*{ext.upper()}'))])

            images = sorted(images, key=lambda k: natural_key(k))
            labels = sorted(labels, key=lambda k: natural_key(k))
            images_and_targets.extend([{'image': str(image), 'label': str(label)} for image, label in zip(images, labels)])

        else:
            for ext in IMG_EXTENSIONS:
                images_and_targets.extend([{'image': str(file), 'label': None}
                                           for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}'))])
            images_and_targets = sorted(images_and_targets, key=lambda k: natural_key(k['image']))

        return images_and_targets

    def load_id_mapping(self):
        root_path = Path(self.conf_data.path.root)

        if isinstance(self.conf_data.id_mapping, DictConfig):
            return dict(self.conf_data.id_mapping)
        elif isinstance(self.conf_data.id_mapping, ListConfig):
            return dict(enumerate(self.conf_data.id_mapping))
        elif isinstance(self.conf_data.id_mapping, str):
            id_mapping_path = root_path / self.conf_data.id_mapping
            if not os.path.exists(id_mapping_path):
                raise FileNotFoundError(f"File not found: {id_mapping_path}")

            with open(id_mapping_path, 'r') as f:
                id_mapping = json.load(f)
            if isinstance(id_mapping, list):
                id_mapping = dict(enumerate(id_mapping))

            return id_mapping
        else:
            raise ValueError(f"Unknown type for id_mapping! {type(self.conf_data.id_mapping)}")

    def load_class_map(self, id_mapping):
        idx_to_class: Dict[int, str] = {}
        label_value_to_idx: Dict[Union[int, Tuple], int] = {}
        for class_idx, (label_value, class_name) in enumerate(id_mapping.items()):
            assert isinstance(class_name, str), "You need to format id_mapping with key for class_index, value for class_name."
            if isinstance(label_value, (int, tuple)):
                idx_to_class[class_idx] = class_name
                label_value_to_idx[label_value] = class_idx
                continue

            # Check tuple string
            assert isinstance(label_value, str) and label_value.strip().startswith("(") and label_value.strip().endswith(")"), \
                f"Unknown type for color index! Should be one of (int, tuple-style str, tuple)... but {type(label_value)}"
            label_value_tuple: Tuple[int, int, int] = as_tuple(label_value)
            idx_to_class[class_idx] = class_name
            label_value_to_idx[label_value_tuple] = class_idx

        return {'idx_to_class': idx_to_class, 'label_value_to_idx': label_value_to_idx}

    def load_huggingface_samples(self):
        from datasets import load_dataset

        cache_dir = self.conf_data.metadata.custom_cache_dir
        root = self.conf_data.metadata.repo
        subset_name = self.conf_data.metadata.subset
        if cache_dir is not None:
            cache_dir = Path(cache_dir)
            Path(cache_dir).mkdir(exist_ok=True, parents=True)
        total_dataset = load_dataset(root, name=subset_name, cache_dir=cache_dir)

        assert isinstance(self.conf_data.id_mapping, (ListConfig, DictConfig))
        if isinstance(self.conf_data.id_mapping, ListConfig):
            self.conf_data.id_mapping = dict(enumerate(self.conf_data.id_mapping))

        misc = self.load_class_map(id_mapping=self.conf_data.id_mapping)

        exists_valid = 'validation' in total_dataset
        exists_test = 'test' in total_dataset

        train_samples = total_dataset['train']
        valid_samples = None
        if exists_valid:
            valid_samples = total_dataset['validation']
        test_samples = None
        if exists_test:
            test_samples = total_dataset['test']

        if not exists_valid:
            splitted_datasets = train_samples.train_test_split(test_size=(1 - self.train_valid_split_ratio))
            train_samples = splitted_datasets['train']
            valid_samples = splitted_datasets['test']
        return train_samples, valid_samples, test_samples, misc


class SegmentationCustomDataset(BaseCustomDataset):

    def __init__(self, conf_data, conf_augmentation, model_name, idx_to_class,
                 split, samples, transform=None, **kwargs):
        super(SegmentationCustomDataset, self).__init__(
            conf_data, conf_augmentation, model_name, idx_to_class,
            split, samples, transform, **kwargs
        )
        assert "label_value_to_idx" in kwargs
        self.label_value_to_idx = kwargs["label_value_to_idx"]

        self.label_image_mode: Literal['RGB', 'L', 'P'] = str(conf_data.label_image_mode).upper() \
            if conf_data.label_image_mode is not None else 'L'

    def cache_dataset(self):
        logger.info(f'Caching | Loading samples of {self.mode} to memory... This can take minutes.')

        def _load(i, samples):
            image = Image.open(Path(samples[i]['image'])).convert('RGB')
            label = self.samples[i]['label']
            if label is not None:
                label = Image.open(Path(label)).convert(self.label_image_mode)
            return i, image, label

        num_threads = 8 # TODO: Compute appropriate num_threads
        all_indices = list(range(len(self.samples)))
        load_imgs = ThreadPool(num_threads).imap(
            partial(_load, samples=self.samples),
            all_indices
        )
        for i, image, label in load_imgs:
            self.samples[i]['image'] = image
            self.samples[i]['label'] = label

        self.cache = True

    def __getitem__(self, index):
        if self.cache:
            img = self.samples[index]['image']
            label = self.samples[index]['label']
        else:
            img_path = self.samples[index]['image']
            ann_path = self.samples[index]['label']
            img = Image.open(Path(img_path)).convert('RGB')
            label = Image.open(Path(ann_path)) if ann_path is not None else None

        w, h = img.size

        outputs = {}
        outputs.update({'indices': index})
        if label is None:
            out = self.transform(image=img)
            outputs.update({'pixel_values': out['image'], 'org_shape': (h, w)})
            return outputs

        if self.label_image_mode == 'L':
            mask = label
        else: # RGB, P
            label = label.convert('RGB')
            label = np.array(label)
            mask = np.zeros((label.shape[0], label.shape[1]), dtype=np.uint8) + 255 # Set undefined as 255
            for label_value in self.label_value_to_idx:
                class_mask = (label == np.array(label_value)).all(axis=-1)
                mask[class_mask] = self.label_value_to_idx[label_value]
            mask = Image.fromarray(mask, mode='L')

        if 'pidnet' in self.model_name:
            edge = generate_edge(np.array(mask))
            out = self.transform(image=img, mask=mask, edge=edge, dataset=self)
            outputs.update({'pixel_values': out['image'], 'labels': out['mask'], 'edges': out['edge'].float()})
        else:
            out = self.transform(image=img, mask=mask, dataset=self)
            outputs.update({'pixel_values': out['image'], 'labels': out['mask']})

        if self._split in ['train', 'training']:
            return outputs

        assert self._split in ['val', 'valid', 'test']
        # outputs.update({'org_img': org_img, 'org_shape': (h, w)})  # TODO: return org_img with batch_size > 1
        outputs.update({'org_shape': (h, w)})
        return outputs


class SegmentationHFDataset(BaseHFDataset):

    def __init__(
            self,
            conf_data,
            conf_augmentation,
            model_name,
            idx_to_class,
            split,
            huggingface_dataset,
            transform=None,
            **kwargs
    ):
        root = conf_data.metadata.repo
        super(SegmentationHFDataset, self).__init__(
            conf_data,
            conf_augmentation,
            model_name,
            root,
            split,
            transform,
        )

        self.idx_to_class = idx_to_class
        self.samples = huggingface_dataset

        assert "label_value_to_idx" in kwargs
        self.label_value_to_idx = kwargs["label_value_to_idx"]

        self.label_image_mode: Literal['RGB', 'L', 'P'] = str(conf_data.label_image_mode).upper() \
            if conf_data.label_image_mode is not None else 'L'

        self.image_feature_name = conf_data.metadata.features.image
        self.label_feature_name = conf_data.metadata.features.label

    @property
    def num_classes(self):
        return len(self.idx_to_class)

    @property
    def class_map(self):
        return self.idx_to_class

    def __len__(self):
        return self.samples.num_rows

    def __getitem__(self, index):

        img_name = f"{index:06d}"
        img: Image.Image = self.samples[index][self.image_feature_name]
        label: Image.Image = self.samples[index][self.label_feature_name] if self.label_feature_name in self.samples[index] else None

        if self.label_image_mode == 'L':
            mask = label
        else: # RGB, P
            label = label.convert('RGB')
            label = np.array(label)
            mask = np.zeros((label.shape[0], label.shape[1]), dtype=np.uint8) + 255 # Set undefined as 255
            for label_value in self.label_value_to_idx:
                class_mask = (label == np.array(label_value)).all(axis=-1)
                mask[class_mask] = self.label_value_to_idx[label_value]
            mask = Image.fromarray(mask, mode='L')

        w, h = img.size

        if label is None:
            out = self.transform(image=img)
            return {'pixel_values': out['image'], 'name': img_name, 'org_shape': (h, w)}

        outputs = {}

        if 'pidnet' in self.model_name:
            edge = generate_edge(np.array(label))
            out = self.transform(image=img, mask=mask, edge=edge, dataset=self)
            outputs.update({'pixel_values': out['image'], 'labels': out['mask'], 'edges': out['edge'].float(), 'name': img_name})
        else:
            out = self.transform(image=img, mask=mask, dataset=self)
            outputs.update({'pixel_values': out['image'], 'labels': out['mask'], 'name': img_name})

        outputs.update({'indices': index})
        if self._split in ['train', 'training']:
            return outputs

        assert self._split in ['val', 'valid', 'test']
        # outputs.update({'org_img': org_img, 'org_shape': (h, w)})  # TODO: return org_img with batch_size > 1
        outputs.update({'org_shape': (h, w)})
        return outputs

    def cache_dataset(self):
        raise NotImplementedError("Caching is not implemented for Hugging Face datasets yet.")



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
EfficientFormer implementation
"""
import copy
import itertools
import math
import os
from typing import Dict, Optional, List

from omegaconf import DictConfig
import torch
import torch.nn as nn

from ...op.base_metaformer import (
    ChannelMLP,
    Image2Sequence,
    MetaFormer,
    MetaFormerBlock,
    MetaFormerEncoder,
    MultiHeadAttention,
    Pooling,
)
from ...op.custom import ConvLayer
from ...op.depth import DropPath
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ['efficientformer']

SUPPORTING_TASK = ['classification', 'segmentation', 'detection', 'pose_estimation']


class EfficientFormerStem(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.stem = nn.Sequential(
            ConvLayer(in_channels, out_channels // 2, kernel_size=3, stride=2, padding=1, bias=True,
                      use_act=True, use_norm=True, norm_type='batch_norm', act_type='relu'),
            ConvLayer(out_channels // 2, out_channels, kernel_size=3, stride=2, padding=1, bias=True,
                      use_act=True, use_norm=True, norm_type='batch_norm', act_type='relu'),
        )

    def forward(self, x):
        x = self.stem(x)
        return x


class EfficientFormerEmbedding(nn.Module):
    def __init__(self, in_channels, out_channels, patch_size, stride, padding):
        super().__init__()
        self.proj = ConvLayer(
            in_channels, out_channels, kernel_size=patch_size, stride=stride, padding=padding, bias=True,
            use_norm=True, use_act=False, norm_type='batch_norm'
        )

    def forward(self, x):
        x = self.proj(x)  # B x C x H//stride x W//stride
        return x


class EfficientFormerMeta4DMLP(nn.Module):
    def __init__(self, hidden_size, intermediate_size=None, hidden_dropout_prob=0., hidden_activation_type='gelu'):
        # hidden_size: in_features
        # intermediate_size: hidden_features
        # hidden_activation_type = 'gelu'
        # hidden_dropout_prob = drop = 0.
        super().__init__()
        intermediate_size = intermediate_size or hidden_size

        self.ffn = nn.Sequential()
        self.ffn.add_module(
            'conv_1x1_1', ConvLayer(hidden_size, intermediate_size, kernel_size=1, bias=True,
                                    use_norm=True, use_act=True, norm_type='batch_norm',
                                    act_type=hidden_activation_type)
        )
        self.ffn.add_module('drop', nn.Dropout(p=hidden_dropout_prob))
        self.ffn.add_module(
            'conv_1x1_2', ConvLayer(intermediate_size, hidden_size, kernel_size=1, bias=True,
                                    use_norm=True, use_act=False, norm_type='batch_norm')
        )

        self.dropout = nn.Dropout(p=hidden_dropout_prob)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Conv2d):
            torch.nn.init.trunc_normal_(m.weight, std=.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.ffn(x)
        x = self.dropout(x)
        return x


class Meta3D(MetaFormerBlock):
    def __init__(self, hidden_size, num_attention_heads, attention_hidden_size,
                 attention_dropout_prob, attention_ratio, attention_bias_resolution,
                 intermediate_ratio, hidden_dropout_prob, hidden_activation_type='gelu',
                 layer_norm_eps=1e-5,
                 drop_path=0., use_layer_scale=True, layer_scale_init_value=1e-5):

        super().__init__(hidden_size, layer_norm_eps)
        self.layernorm_before = nn.LayerNorm(hidden_size, eps=layer_norm_eps)

        # intermediate_ratio: mlp_ratio
        # hidden_activation_type: act_type
        # hidden_dropout_prob: drop
        # hidden_size: dim
        # hidden_size: key_dim * num_attention_heads

        # self.token_mixer = Attention(dim)
        # (dim=384, key_dim=32, num_heads=8, attn_ratio=4, resolution=16)
        # num_attention_heads: 8
        # hidden_size: key_dim*num_attention_heads = 32 * 8
        # value_hidden_size: key_dim*num_attention_heads*attn_ratio = 32 * 8 * 4
        # attention_bias_resolution: resolution = 16
        self.token_mixer = MultiHeadAttention(
            hidden_size, num_attention_heads,
            attention_hidden_size=attention_hidden_size,
            attention_dropout_prob=attention_dropout_prob,
            value_hidden_size=attention_hidden_size*attention_ratio,
            use_attention_bias=True,
            attention_bias_resolution=attention_bias_resolution
        )
        self.layernorm_after = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        intermediate_size = int(hidden_size * intermediate_ratio)
        self.channel_mlp = ChannelMLP(
            hidden_size, intermediate_size, hidden_dropout_prob, hidden_activation_type
        )

        # The following two techniques are useful to train deep PoolFormers.
        self.drop_path = DropPath(drop_path) if drop_path > 0. \
            else nn.Identity()
        self.use_layer_scale = use_layer_scale
        if use_layer_scale:
            self.layer_scale_1 = nn.Parameter(
                layer_scale_init_value * torch.ones((hidden_size)), requires_grad=True)
            self.layer_scale_2 = nn.Parameter(
                layer_scale_init_value * torch.ones((hidden_size)), requires_grad=True)

    def forward(self, x):
        if self.use_layer_scale:
            x = x + self.drop_path(
                self.layer_scale_1.unsqueeze(0).unsqueeze(0) *
                self.token_mixer(self.layernorm_before(x))
            )
            x = x + self.drop_path(
                self.layer_scale_2.unsqueeze(0).unsqueeze(0) *
                self.channel_mlp(self.layernorm_after(x))
            )

        else:
            x = x + self.drop_path(self.token_mixer(self.layernorm_before(x)))
            x = x + self.drop_path(self.channel_mlp(self.layernorm_after(x)))
        return x


class Meta4D(MetaFormerBlock):
    def __init__(self, hidden_size, pool_size=3, intermediate_ratio=4.,
                 hidden_dropout_prob=0., hidden_activation_type='gelu',
                 drop_path=0., use_layer_scale=True, layer_scale_init_value=1e-5):
        # intermediate_ratio: mlp_ratio
        # hidden_size: dim
        # hidden_dropout_prob: drop
        # hidden_activation_type: act_type
        super().__init__(hidden_size, 0.)
        self.layernorm_before = nn.Identity()  # not used
        self.layernorm_after = nn.Identity()  # not used
        self.token_mixer = Pooling(pool_size)
        intermediate_size = int(hidden_size * intermediate_ratio)
        self.channel_mlp = EfficientFormerMeta4DMLP(
            hidden_size, intermediate_size, hidden_dropout_prob, hidden_activation_type
        )

        self.drop_path = DropPath(drop_path) if drop_path > 0. \
            else nn.Identity()
        self.use_layer_scale = use_layer_scale
        if use_layer_scale:
            self.layer_scale_1 = nn.Parameter(
                layer_scale_init_value * torch.ones((hidden_size)), requires_grad=True)
            self.layer_scale_2 = nn.Parameter(
                layer_scale_init_value * torch.ones((hidden_size)), requires_grad=True)

    def forward(self, x):
        if self.use_layer_scale:
            x = x + self.drop_path(
                self.layer_scale_1.unsqueeze(-1).unsqueeze(-1) * self.token_mixer(x)
            )
            x = x + self.drop_path(
                self.layer_scale_2.unsqueeze(-1).unsqueeze(-1) * self.channel_mlp(x)
            )
        else:
            x = x + self.drop_path(self.token_mixer(x))
            x = x + self.drop_path(self.channel_mlp(x))
        return x


class EfficientFormerEncoder(MetaFormerEncoder):
    def __init__(self, use_intermediate_features, num_blocks, hidden_sizes,
                 num_attention_heads, attention_hidden_size, attention_dropout_prob,
                 attention_ratio, attention_bias_resolution,
                 pool_size, intermediate_ratio, hidden_dropout_prob, hidden_activation_type,
                 layer_norm_eps,
                 drop_path_rate, use_layer_scale, layer_scale_init_value,
                 downsamples, down_patch_size, down_stride, down_pad,
                 vit_num=1):
        # intermediate_ratio: mlp_ratio
        # hidden_size: dim
        # hidden_dropout_prob: drop
        # hidden_activation_type: act_type

        # intermediate_ratio: mlp_ratio
        # hidden_activation_type: act_type
        # hidden_dropout_prob: drop
        # hidden_size: dim
        # attention_hidden_size: key_dim * num_attention_heads

        # self.token_mixer = Attention(dim)
        # (dim=384, key_dim=32, num_heads=8, attn_ratio=4, resolution=16)
        # num_attention_heads: 8
        # hidden_size: key_dim*num_attention_heads = 32 * 8
        # value_hidden_size: key_dim*num_attention_heads*attn_ratio = 32 * 8 * 4
        # attention_bias_resolution: resolution = 16

        super().__init__()
        self.use_intermediate_features = use_intermediate_features

        blocks = []
        for module_idx in range(len(num_blocks)):
            stage = self.meta_blocks(
                num_blocks, module_idx, hidden_sizes[module_idx],
                num_attention_heads, attention_hidden_size, attention_dropout_prob,
                attention_ratio, attention_bias_resolution,
                pool_size, intermediate_ratio, hidden_dropout_prob, hidden_activation_type,
                layer_norm_eps,
                drop_path_rate, use_layer_scale, layer_scale_init_value,
                vit_num
            )
            blocks.append(stage)
            if module_idx >= len(num_blocks) - 1:
                break
            if downsamples[module_idx] or hidden_sizes[module_idx] != hidden_sizes[module_idx + 1]:
                # downsampling between two stages
                blocks.append(
                    EfficientFormerEmbedding(
                        hidden_sizes[module_idx],
                        hidden_sizes[module_idx + 1],
                        down_patch_size,
                        down_stride,
                        down_pad
                    )
                )

        self.blocks = nn.ModuleList(blocks)

        if self.use_intermediate_features:
            # add a norm layer for each output
            self.intermediate_features_indices = [0, 2, 4, 6]
            for i_emb, i_layer in enumerate(self.intermediate_features_indices):
                if i_emb == 0 and os.environ.get('FORK_LAST3', None):
                    layer = nn.Identity()
                else:
                    layer = nn.GroupNorm(1, hidden_sizes[i_emb])
                self.add_module(f'norm{i_layer}', layer)

    @staticmethod
    def meta_blocks(num_blocks, module_idx, hidden_size,
                    num_attention_heads, attention_hidden_size, attention_dropout_prob,
                    attention_ratio, attention_bias_resolution,
                    pool_size, intermediate_ratio, hidden_dropout_prob, hidden_activation_type,
                    layer_norm_eps,
                    drop_path_rate, use_layer_scale, layer_scale_init_value,
                    vit_num=1):

        blocks = []
        if module_idx == 3 and vit_num == num_blocks[module_idx]:
            blocks.append(Image2Sequence())
        for block_idx in range(num_blocks[module_idx]):
            block_dpr = drop_path_rate * (block_idx + sum(num_blocks[:module_idx])) / (sum(num_blocks) - 1)
            if module_idx == 3 and num_blocks[module_idx] - block_idx <= vit_num:
                blocks.append(Meta3D(
                    hidden_size, num_attention_heads, attention_hidden_size,
                    attention_dropout_prob,
                    attention_ratio,
                    attention_bias_resolution,
                    intermediate_ratio,
                    hidden_dropout_prob,
                    hidden_activation_type,
                    layer_norm_eps,
                    block_dpr,
                    use_layer_scale,
                    layer_scale_init_value
                ))
            else:
                blocks.append(Meta4D(
                    hidden_size, pool_size, intermediate_ratio,
                    hidden_dropout_prob,
                    hidden_activation_type,
                    block_dpr,
                    use_layer_scale,
                    layer_scale_init_value
                ))
                if module_idx == 3 and num_blocks[module_idx] - block_idx - 1 == vit_num:
                    blocks.append(Image2Sequence())

        blocks = nn.Sequential(*blocks)
        return blocks

    def forward(self, x):
        all_hidden_states = () if self.use_intermediate_features else None
        for idx, block in enumerate(self.blocks):
            # if len(x.size()) == 4:
            B, C, H, W = x.size()
            x = block(x)
            if self.use_intermediate_features and idx in self.intermediate_features_indices:
                norm_layer = getattr(self, f'norm{idx}')
                # @deepkyu: [fx tracing] len(x.size()) == 3 when idx == 6 (with Meta3D, in efficientformer_l1)
                # if len(x.size()) != 4:
                if idx == self.intermediate_features_indices[-1]:
                    x = x.transpose(1, 2).reshape(B, C, H, W)
                x = norm_layer(x)
                all_hidden_states = all_hidden_states + (x,)

        if self.use_intermediate_features:
            # output the features of four stages for dense prediction
            return all_hidden_states
        # output only the features of last layer for image classification
        return x


class EfficientFormer(MetaFormer):

    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        # Check task compatibilty
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f'EfficientFormer is not supported on {self.task} task now.'
        self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST

        # Get params
        num_blocks = [stage.num_blocks for stage in stage_params]
        hidden_sizes = [stage.channels for stage in stage_params]

        num_attention_heads = params.num_attention_heads
        attention_hidden_size = params.attention_channels
        attention_dropout_prob = params.attention_dropout_prob
        attention_ratio = params.attention_value_expansion_ratio
        intermediate_ratio = params.ffn_intermediate_ratio
        hidden_dropout_prob = params.ffn_dropout_prob
        hidden_activation_type = params.ffn_act_type
        vit_num = params.vit_num

        # Fix as constant
        layer_norm_eps = 1e-5
        layer_scale_init_value = 1e-5
        down_patch_size = 3
        down_stride = 2
        down_pad = 1
        pool_size = 3
        downsamples = [True for _ in stage_params]
        use_layer_scale = True
        attention_bias_resolution = 16
        drop_path_rate = 0.

        super().__init__(hidden_sizes)

        image_channels = 3
        self.patch_embed = EfficientFormerStem(in_channels=image_channels, out_channels=hidden_sizes[0])

        self.encoder = EfficientFormerEncoder(
            self.use_intermediate_features, num_blocks, hidden_sizes,
            num_attention_heads, attention_hidden_size, attention_dropout_prob,
            attention_ratio, attention_bias_resolution,
            pool_size, intermediate_ratio, hidden_dropout_prob, hidden_activation_type,
            layer_norm_eps,
            drop_path_rate, use_layer_scale, layer_scale_init_value,
            downsamples, down_patch_size, down_stride, down_pad,
            vit_num=vit_num
        )

        self.norm = nn.LayerNorm(hidden_sizes[-1], eps=layer_norm_eps)

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK

    def forward(self, x):
        x = self.patch_embed(x)
        x = self.encoder(x)
        if self.use_intermediate_features:
            all_hidden_states = x  # (features)
            return BackboneOutput(intermediate_features=all_hidden_states)
        x = self.norm(x)  # B x N x C
        feat = torch.mean(x, dim=-2)
        return BackboneOutput(last_feature=feat)


def efficientformer(task, conf_model_backbone) -> EfficientFormer:
    return EfficientFormer(task, conf_model_backbone.params, conf_model_backbone.stage_params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Any, Dict

from .base import MetricFactory
from .registry import METRIC_ADAPTORS, METRIC_LIST, PHASE_LIST, TASK_AVAILABLE_METRICS, TASK_DEFAULT_METRICS


def build_metrics(task: str, model_conf, metrics_conf, num_classes, **kwargs) -> MetricFactory:
    metric_names = metrics_conf.metric_names
    classwise_analysis = metrics_conf.classwise_analysis

    if metric_names is None:
        metric_names = TASK_DEFAULT_METRICS[task]
    metric_names = [m.lower() for m in metric_names]
    assert all(metric in TASK_AVAILABLE_METRICS[task] for metric in metric_names), \
        f"Available metrics for {task} are {TASK_AVAILABLE_METRICS[task]}"

    # TODO: This code assumes there is only one loss module. Fix here later.
    if hasattr(model_conf.losses[0], 'ignore_index'):
        kwargs['ignore_index'] = model_conf.losses[0].ignore_index

    metrics = {}
    for phase in PHASE_LIST:
        if phase == 'valid': # classwise_analysis is only available in valid phase
            metrics[phase] = [METRIC_LIST[name](num_classes=num_classes, classwise_analysis=classwise_analysis, **kwargs) for name in metric_names]
        else:
            metrics[phase] = [METRIC_LIST[name](num_classes=num_classes, classwise_analysis=False, **kwargs) for name in metric_names]

    metric_adaptor = METRIC_ADAPTORS[task](metric_names)

    metric_handler = MetricFactory(task, metrics, metric_adaptor, classwise_analysis)
    return metric_handler
